REM
REM  This file is part of the Grumpy project
REM  http://www.grumpy-server.net/
REM
REM  GetMIME takes one string parameter (the path/name of the file to check out) and outputs its
REM  MIME type (if any matched).
REM
REM  Usefull ressource: http://www.iana.org/assignments/media-types
REM

PRIVATE FUNCTION IsStartingBy(ContainerString AS STRING, StartString AS STRING) AS BYTE
  REM Checks if ContainerString is starting by StartString. Returns 1 if so, and 0 otherwise.
  IF LEFT(ContainerString, LEN(StartString)) = StartString THEN
      RETURN 1
    ELSE
      RETURN 0
  END IF
END FUNCTION


FUNCTION GetMIME(FileToCheck AS STRING) AS STRING
  DIM AS INTEGER FileHandler, FileSize, LastDotInFilename, x, y, z
  DIM AS STRING Result, DataBuffer, FileExtension = ""
  LastDotInFilename = INSTRREV(FileToCheck, ".")
  IF LastDotInFilename > 0 THEN FileExtension = UCASE(MID(FileToCheck, LastDotInFilename + 1))
  REM Check wheter a user-defined MIME type is set
  IF LEN(FileExtension) > 0 THEN
    Result = ReadCFG(ConfigFile, "AddMimeExt:" + FileExtension)
  END IF

  IF LEN(Result) = 0 AND VAL(ReadCFG(ConfigFile, "AutoMIME")) = 1 THEN
  REM If no user-definied match, and AutoMIME enabled, then...
    FileHandler = FREEFILE
    IF OPEN(FileToCheck, FOR BINARY, AS #FileHandler) = 0 THEN
      FileSize = LOF(FileHandler)
      REM  *******************************
      REM  *  Here start all signatures  *
      REM  *******************************

      ' 7-zip archive
      IF LEN(Result) = 0 THEN
        IF LOF(FileHandler) > 16 AND FileExtension = "7Z" THEN
          DataBuffer = SPACE(6)
          GET #FileHandler, 1, DataBuffer
          IF DataBuffer = CHR(&h37,&h7A,&hBC,&hAF,&h27,&h1C) THEN Result = "application/x-7z-compressed"
        END IF
      END IF

      ' AU Sun Microsystems audio file
      IF LEN(Result) = 0 THEN
        IF FileSize > 16 THEN
          DataBuffer = SPACE(4)
          GET #FileHandler, 1, DataBuffer
          IF DataBuffer = ".snd" THEN Result = "audio/basic"
        END IF
      END IF

      ' BMP image
      IF LEN(Result) = 0 THEN
        IF FileSize > 16 AND FileExtension = "BMP" THEN
          DataBuffer = SPACE(2)
          GET #FileHandler, 1, DataBuffer
          IF DataBuffer = "BM" THEN Result = "image/x-ms-bmp"  ' Only MS format
        END IF
      END IF

      ' Bzip2 archive
      IF LEN(Result) = 0 THEN
        IF FileSize > 16 THEN
          DataBuffer = SPACE(3)
          GET #FileHandler, 1, DataBuffer
          IF DataBuffer = "BZh" THEN Result = "application/x-bzip"
        END IF
      END IF

      ' DEB package (DEBIAN software)
      IF LEN(Result) = 0 THEN
        IF FileSize > 16 THEN
          DataBuffer = SPACE(22)
          GET #FileHandler, 1, DataBuffer
          IF DataBuffer = "!<arch>" + CHR(&h0A) + "debian-binary " THEN
            DataBuffer = SPACE(14)
            GET #FileHandler, 73, DataBuffer
            IF DataBuffer = "control.tar.gz" THEN Result = "application/x-deb"
          END IF
        END IF
      END IF

      ' EPS (Adobe Encapsulated PostScript)
      IF LEN(Result) = 0 THEN
        IF FileSize > 16 AND FileExtension = "EPS" THEN
          DataBuffer = SPACE(4)
          GET #FileHandler, 1, DataBuffer
          IF DataBuffer = CHR(&hC5,&hD0,&hD3,&hC6) THEN ' EPS containing TIFF/WMF preview
              Result = "image/x-eps"
            ELSE
              DataBuffer = SPACE(11)
              GET #FileHandler, 1, DataBuffer  ' EPS without any preview
              IF DataBuffer = "%!PS-Adobe-" THEN
                DataBuffer = SPACE(10)
                GET #FileHandler, 15, DataBuffer  ' EPS without any preview
                IF INSTR(DataBuffer, " EPSF-") > 0 THEN Result = "image/x-eps"
              END IF
          END IF
        END IF
      END IF

      ' Famicom Disk System dump
      IF LEN(Result) = 0 THEN
        IF FileSize > 32768 AND FileSize < 262144 THEN ' A FDS may be 1-sided (64K) or double-sided (128K) nothing else.
          DataBuffer = SPACE(5)
          GET #FileHandler, 1, DataBuffer ' 4 first bytes are a magic header, while the 5th is the number of sides of the disk (can't be anything else than 1 or 2)
          IF DataBuffer = CHR(&h46,&h44,&h53,&h1A,&h01) OR DataBuffer = CHR(&h46,&h44,&h53,&h1A,&h02) THEN Result = "application/x-nes-disk"
        END IF
      END IF

      ' FLAC audio file (Free Lossless Audio Codec)
      IF LEN(Result) = 0 THEN
        IF FileSize > 16 THEN
          DataBuffer = SPACE(8)
          GET #FileHandler, 1, DataBuffer
          IF DataBuffer = CHR(&h66,&h4C,&h61,&h43,&h00,&h00,&h00,&h22) THEN Result = "audio/x-flac"
        END IF
      END IF

      ' Flash video (FLV)
      IF LEN(Result) = 0 THEN
        IF FileSize > 16 THEN
          DataBuffer = SPACE(4)
          GET #FileHandler, 1, DataBuffer ' The header is "FLV" + version number (v1 is the only one so far)
          IF DataBuffer = "FLV" + CHR(1) THEN Result = "video/x-flv"
        END IF
      END IF

      ' GIF image
      IF LEN(Result) = 0 THEN
        IF FileSize > 16 THEN
          DataBuffer = SPACE(6)
          GET #FileHandler, 1, DataBuffer
          IF DataBuffer = "GIF89a" OR DataBuffer = "GIF87a" THEN Result = "image/gif"
        END IF
      END IF

      ' GZip archive
      IF LEN(Result) = 0 THEN
        IF FileSize > 16 THEN
          DataBuffer = SPACE(3)
          GET #FileHandler, 1, DataBuffer
          IF DataBuffer = CHR(&h1F,&h8B,&h08) THEN Result = "application/x-gzip"
        END IF
      END IF

      ' ICO icon
      IF LEN(Result) = 0 THEN
        IF FileSize > 16 THEN
          DataBuffer = SPACE(4)
          GET #FileHandler, 1, DataBuffer
          IF DataBuffer = CHR(0,0,1,0) THEN Result = "image/vnd.microsoft.icon"
        END IF
      END IF

      ' iNES ROM
      IF LEN(Result) = 0 THEN
        IF FileSize > 16384 AND FileSize < 8*1024*1024 THEN ' Minimum is 16KiB, and can't be technically more than 6MiB (but accepting 8MiB here is no problem anyway)
          DataBuffer = SPACE(4)
          GET #FileHandler, 1, DataBuffer
          IF DataBuffer = "NES" + CHR(&h1A) THEN
            DataBuffer = SPACE(2)
            GET #FileHandler, 5, DataBuffer
            x = DataBuffer[0]  ' number of 16k PRG blocks
            y = DataBuffer[1]  ' number of 8k CHR blocks
            z = 16 + (x * 16384) + (y * 8192) ' Predicted minimum size (may be 128 byte bigger if a title is provided)
            IF FileSize - z >= 0 AND FileSize - z <= 128 THEN
              Result = "application/x-nes-rom"  ' A NES ROM is a 16byte header + x 16K PRG blocks + y 8K CHR blocks + an optional 128 byte description
            END IF
          END IF
        END IF
      END IF

      ' IT music module
      IF LEN(Result) = 0 THEN
        IF FileSize > 64 AND FileExtension = "IT" THEN ' The header itself is bigger than 64 bytes, so the file can't be smaller
          DataBuffer = SPACE(4)
          GET #FileHandler, 1, DataBuffer
          IF DataBuffer = "IMPM" OR DataBuffer = "IMPI" THEN Result = "audio/it"
        END IF
      END IF

      ' JAR Java archive
      IF LEN(Result) = 0 THEN
        IF FileSize > 16 AND FileExtension = "JAR" THEN ' Warning: it's in fact a bare ZIP file...
          DataBuffer = SPACE(10)
          GET #FileHandler, 1, DataBuffer
          IF DataBuffer = CHR(&h50,&h4B,&h03,&h04,&h14,&h00,&h08,&h00,&h08,&h00) THEN Result = "application/java-archive"
        END IF
      END IF

      ' Java class
      IF LEN(Result) = 0 THEN
        IF FileSize > 16 THEN
          DataBuffer = SPACE(4)
          GET #FileHandler, 1, DataBuffer
          IF DataBuffer = CHR(&hca,&hfe,&hba,&hbe) OR DataBuffer = CHR(&hca,&hfe,&hd0,&h0d) THEN Result = "application/x-java-vm"
        END IF
      END IF

      ' JPEG image
      IF LEN(Result) = 0 THEN
        IF FileSize > 64 THEN ' The smallest JPEG file I could get with GIMP was 284 bytes
          DataBuffer = SPACE(3)       ' long, therefore 64 bytes seems like a very reasonable minimum.
          GET #FileHandler, 1, DataBuffer  ' The fourth byte is variable (E0, E1, E2, etc...), so ignore it.
          IF DataBuffer = CHR(&hFF,&hD8,&hFF) THEN
            DataBuffer = SPACE(2)
            GET #FileHandler, LOF(FileHandler) - 1, DataBuffer
            IF DataBuffer = CHR(&hFF,&hD9) THEN ' Check out the JPEG trailer (should be FF D9)
                Result = "image/jpeg"
              ELSE
                REM Some (corrupted?) JPEGs are adding few nul chars after the trailer...
                x = 31 : y = 0
                DataBuffer = SPACE(32)
                GET #FileHandler, LOF(FileHandler) - 31, DataBuffer
                WHILE x > 0 AND y = 0
                  IF DataBuffer[x] > 0 THEN
                    IF MID(DataBuffer, x, 2) = CHR(&hFF,&hD9) THEN Result = "image/jpeg"
                    y = 1
                  END IF
                  x -= 1
                WEND
            END IF
          END IF
        END IF
      END IF

      ' MIDI music
      IF LEN(Result) = 0 THEN
        IF FileSize > 16 THEN
          DataBuffer = SPACE(4)
          GET #FileHandler, 1, DataBuffer
          IF DataBuffer = "MThd" THEN Result = "audio/midi"
        END IF
      END IF

      ' MS-Office documents (old types)
      IF LEN(Result) = 0 THEN
        IF FileSize > 514 THEN
          DataBuffer = SPACE(8)
          GET #FileHandler, 1, DataBuffer
          IF DataBuffer = CHR(&hD0,&hCF,&h11,&hE0,&hA1,&hB1,&h1A,&hE1) THEN
            DataBuffer = SPACE(1) ' Retrieve the MSOFFICE subheader byte
            GET #FileHandler, 514, DataBuffer
            SELECT CASE DataBuffer
              CASE CHR(&hA5) ' Word
                Result = "application/msword"
              CASE CHR(&h08) ' Excel
                Result = "application/vnd.ms-excel"
              CASE CHR(&h00) ' Powerpoint
                Result = "application/vnd.ms-powerpoint"
            END SELECT
          END IF
        END IF
      END IF

      ' MS-Office 2007/2008 documents
      IF LEN(Result) = 0 THEN
        IF FileSize > 32 THEN
          DataBuffer = SPACE(8)
          GET #FileHandler, 1, DataBuffer
          IF DataBuffer = CHR(&h50,&h4B,&h03,&h04,&h14,&h00,&h06,&h00) THEN
            SELECT CASE FileExtension
              CASE "DOCX", "DOCM" ' Word
                Result = "application/vnd.openxmlformats-officedocument.wordprocessingml.document"
              CASE "XLSX" ' Excel
                Result = "application/vnd.openxmlformats-officedocument.spreadsheetml.sheet"
              CASE "PPTX" ' Powerpoint
                Result = "application/vnd.openxmlformats-officedocument.presentationml.presentation"
            END SELECT
          END IF
        END IF
      END IF

      ' NES sound file
      IF LEN(Result) = 0 THEN
        IF FileSize > 16 THEN
          DataBuffer = SPACE(6)
          GET #FileHandler, 1, DataBuffer
          IF DataBuffer = CHR(&h4e,&h45,&h53,&h4d,&h1a,&h01) THEN Result = "application/x-nes-sound-file"
        END IF
      END IF

      ' Ogg container
      IF LEN(Result) = 0 THEN
        IF FileSize > 16 THEN
          DataBuffer = SPACE(8)
          GET #FileHandler, 1, DataBuffer
          IF DataBuffer = CHR(&h4f,&h67,&h67,&h53,&h00,&h02,&h00,&h00) THEN Result = "application/ogg"
        END IF
      END IF

      ' OpenOffice document
      IF LEN(Result) = 0 THEN
        IF FileSize > 32 THEN
          DataBuffer = SPACE(2)
          GET #FileHandler, 1, DataBuffer
          IF DataBuffer = "PK" THEN
            DataBuffer = SPACE(43)
            GET #FileHandler, 31, DataBuffer
            IF DataBuffer = "mimetypeapplication/vnd.oasis.opendocument." THEN
              DataBuffer = SPACE(24)
              GET #FileHandler, 74, DataBuffer
              DIM OpenOfficeMimeTypes(1 TO 17) AS STRING => {"presentation-template",_
                                                             "spreadsheet-template",_
                                                             "graphics-template",_
                                                             "formula-template",_
                                                             "chart-template",_
                                                             "image-template",_
                                                             "text-template",_
                                                             "presentation",_
                                                             "text-master",_
                                                             "spreadsheet",_
                                                             "database",_
                                                             "graphics",_
                                                             "text-web",_
                                                             "formula",_
                                                             "chart",_
                                                             "image",_
                                                             "text"}
              x = 1
              WHILE x <= UBOUND(OpenOfficeMimeTypes) AND LEN(Result) = 0
                IF IsStartingBy(DataBuffer, OpenOfficeMimeTypes(x)) = 1 THEN
                    Result = "vnd.oasis.opendocument." + OpenOfficeMimeTypes(x)
                  ELSE
                    x += 1
                END IF
              WEND
            END IF
          END IF
        END IF
      END IF

      ' PCX image
      IF LEN(Result) = 0 THEN
        IF FileSize > 128 THEN ' 128 is the PCX header's size
          DataBuffer = SPACE(1)
          GET #FileHandler, 1, DataBuffer
          IF DataBuffer = CHR(10) THEN  ' ID byte must be 10 (Zsoft PCX)
            GET #FileHandler, 2, DataBuffer
            IF ASC(DataBuffer) <= 5 THEN ' Version byte may be 0, 2, 3, 4 or 5
              GET #FileHandler, 3, DataBuffer
              IF DataBuffer = CHR(1) THEN
                GET #FileHandler, 4, DataBuffer
                IF ASC(DataBuffer) = 1 OR ASC(DataBuffer) = 2 OR ASC(DataBuffer) = 4 OR ASC(DataBuffer) = 8 THEN ' Bit per pixel has to be 1, 2, 4 or 8
                  GET #FileHandler, 65, DataBuffer
                  IF DataBuffer = CHR(0) THEN Result = "image/x-pcx"  ' Reserved byte should be 0
                END IF
              END IF
            END IF
          END IF
        END IF
      END IF

      ' PDF document
      IF LEN(Result) = 0 THEN
        IF FileSize > 16 THEN
          DataBuffer = SPACE(4)
          GET #FileHandler, 1, DataBuffer
          IF DataBuffer = "%PDF" THEN
            DataBuffer = SPACE(7)  ' PDF trailer should be "%%EOF" + LF or "%%EOF" + CR/LF
            GET #FileHandler, LOF(FileHandler) - 7, DataBuffer
            IF INSTR(DataBuffer, "%%EOF") > 0 THEN Result = "application/pdf"
          END IF
        END IF
      END IF

      ' PNG image
      IF LEN(Result) = 0 THEN
        IF FileSize > 16 THEN
          DataBuffer = SPACE(8)
          GET #FileHandler, 1, DataBuffer
          IF DataBuffer = CHR(&h89,&h50,&h4e,&h47,&h0d,&h0a,&h1a,&h0a) THEN Result = "image/png"
        END IF
      END IF

      ' QuickTime movie
      IF LEN(Result) = 0 THEN
        IF FileSize > 16 THEN
          DataBuffer = SPACE(4)
          GET #FileHandler, 5, DataBuffer
          IF DataBuffer = "moov" THEN
              Result = "video/quicktime"
            ELSE
              GET #FileHandler, 37, DataBuffer
              IF DataBuffer = "moov" THEN Result = "video/quicktime"
          END IF
        END IF
      END IF

      ' RAR archive
      IF LEN(Result) = 0 THEN
        IF FileSize > 16 AND FileExtension = "RAR" THEN
          DataBuffer = SPACE(7)
          GET #FileHandler, 1, DataBuffer
          IF DataBuffer = CHR(&h52,&h61,&h72,&h21,&h1A,&h07,&h00) THEN Result = "application/x-rar-compressed"
        END IF
      END IF

      ' RA RealMedia streaming audio file
      IF LEN(Result) = 0 THEN
        IF FileSize > 16 THEN
          DataBuffer = SPACE(5)
          GET #FileHandler, 1, DataBuffer
          IF DataBuffer = CHR(&h2E,&h72,&h61,&hFD,&h00) THEN Result = "audio/x-realaudio"
        END IF
      END IF

      ' RM RealMedia streaming media file
      IF LEN(Result) = 0 THEN
        IF FileSize > 16 THEN
          DataBuffer = SPACE(4)
          GET #FileHandler, 1, DataBuffer
          IF DataBuffer = CHR(&h2E,&h52,&h4D,&h46) THEN Result = "audio/vnd.rn-realmedia"
        END IF
      END IF

      ' RPM package
      IF LEN(Result) = 0 THEN
        IF FileSize > 16 THEN
          DataBuffer = SPACE(4)
          GET #FileHandler, 1, DataBuffer
          IF DataBuffer = CHR(&hED,&hAB,&hEE,&hDB) THEN Result = "application/x-rpm"
        END IF
      END IF

      ' RTF document
      IF LEN(Result) = 0 THEN
        IF FileSize > 16 THEN
          DataBuffer = SPACE(5)
          GET #FileHandler, 1, DataBuffer
          IF DataBuffer = "{\rtf" THEN Result = "text/rtf"
        END IF
      END IF

      ' SWF - Shockwave flash
      IF LEN(Result) = 0 THEN
        IF FileSize > 16 THEN
          DataBuffer = SPACE(3)
          GET #FileHandler, 1, DataBuffer
          IF DataBuffer = "FWS" OR DataBuffer = "CWS" THEN Result = "application/x-shockwave-flash"
        END IF
      END IF

      ' TAR archive
      IF LEN(Result) = 0 THEN
        IF FileSize > 261 AND FileExtension = "TAR" THEN
          DataBuffer = SPACE(5)
          GET #FileHandler, 258, DataBuffer
          IF DataBuffer = "ustar" THEN Result = "application/x-tar"
        END IF
      END IF

      ' TIFF image
      IF LEN(Result) = 0 THEN
        IF FileSize > 16 THEN
          DataBuffer = SPACE(4)
          GET #FileHandler, 1, DataBuffer
          IF DataBuffer = CHR(&h4D,&h4D,&h00,&h00) OR DataBuffer = CHR(&h4D,&h4D,&h00,&h2B) OR DataBuffer = CHR(&h49,&h49,&h2A,&h00) THEN Result = "image/tiff"
        END IF
      END IF

      ' TXT plain text
      IF LEN(Result) = 0 AND FileExtension = "TXT" THEN
        IF FileSize >= 3 THEN
            DataBuffer = SPACE(3)
            GET #FileHandler, 1, DataBuffer
            IF DataBuffer = CHR(&hef,&hbb,&hbf) THEN ' Check for UTF-8 BOM
                Result = "text/plain; charset=UTF-8"
              ELSE
                IF FileSize > 1024 THEN y = 1024 ELSE y = LOF(FileHandler)
                DataBuffer = SPACE(y)
                z = 0
                x = 0
                WHILE x < y AND z = 0   ' Look for CHR(0). If found, then it is NOT a text file
                  IF DataBuffer[x] = 0 THEN z = 1
                  x += 1
                WEND
                IF z = 0 THEN Result = "text/plain"
            END IF
          ELSE
            Result = "text/plain"
        END IF
      END IF

      ' WAVe audio
      IF LEN(Result) = 0 THEN
        IF FileSize > 16 THEN
          DataBuffer = SPACE(4)
          GET #FileHandler, 1, DataBuffer
          IF DataBuffer = "RIFF" THEN
            DataBuffer = SPACE(8)
            GET #FileHandler, 9, DataBuffer
            IF DataBuffer = "WAVEfmt " THEN Result = "audio/x-wav"
          END IF
        END IF
      END IF

      ' ZIP archive
      IF LEN(Result) = 0 THEN
        IF FileSize > 16 AND FileExtension = "ZIP" THEN
          DataBuffer = SPACE(4)
          GET #FileHandler, 1, DataBuffer
          IF DataBuffer = CHR(&h50,&h4b,&h03,&h04) THEN Result = "application/zip"
        END IF
      END IF

      ' ZOO archive
      IF LEN(Result) = 0 THEN
        IF FileSize > 16 AND FileExtension = "ZOO" THEN
          DataBuffer = SPACE(4)
          GET #FileHandler, 1, DataBuffer
          IF DataBuffer = CHR(&h5A,&h4F,&h4F,&h20) THEN Result = "application/x-zoo"
        END IF
      END IF

      REM  ***********************
      REM  *  Signatures ended!  *
      REM  ***********************
      CLOSE #FileHandler
    END IF
  END IF
  RETURN Result
END FUNCTION
